from collections.abc import Callable
from typing import get_type_hints

import pydantic

from liblaf.cherries import config, integration, presets
from liblaf.cherries import pathutils as _path


def run[C: pydantic.BaseModel, T](
    main: Callable[[], T] | Callable[[C], T],
    *,
    play: bool = False,
    preset: presets.Preset = presets.default,
) -> T:
    exp: integration.Experiment = start(preset=preset, play=play)
    type_hints: dict[str, type[C]] = get_type_hints(main)
    del type_hints["return"]
    args: list[C] = []
    if len(type_hints) == 1:
        cls: type[C] = next(iter(type_hints.values()))
        cfg: C = cls()
        args.append(cfg)
        for path in config.get_inputs(cfg):
            exp.log_input(path)
    try:
        result: T = main(*args)
    finally:
        if len(type_hints) == 1:
            cfg: C = args[0]
            for path in config.get_outputs(cfg):
                exp.log_output(path)
        exp.end()
    return result


def start(
    preset: presets.Preset = presets.default, *, play: bool = False
) -> integration.Experiment:
    if play:
        preset = presets.playground
    exp: integration.Experiment = preset(integration.exp)
    exp.start()
    exp.log_other("cherries.entrypoint", _path.entrypoint(absolute=False))
    exp.log_other("cherries.exp-dir", _path.exp_dir(absolute=False))
    exp.log_other("cherries.start-time", exp.start_time)
    return exp


def end() -> None:
    integration.exp.end()



import datetime
import functools
from os import PathLike
from pathlib import Path

import attrs
import comet_ml as comet

from liblaf.cherries import pathutils as _path

from ._abc import (
    AddTag,
    AddTags,
    End,
    LogAsset,
    LogCode,
    LogMetric,
    LogMetrics,
    LogOther,
    LogOthers,
    LogParam,
    LogParams,
    Start,
)


@attrs.define
class Experiment:
    add_tag: AddTag = attrs.field(factory=AddTag, init=False)
    add_tags: AddTags = attrs.field(factory=AddTags, init=False)
    end: End = attrs.field(factory=End, init=False)
    log_asset: LogAsset = attrs.field(factory=LogAsset, init=False)
    log_code: LogCode = attrs.field(factory=LogCode, init=False)
    log_metric: LogMetric = attrs.field(factory=LogMetric, init=False)
    log_metrics: LogMetrics = attrs.field(factory=LogMetrics, init=False)
    log_other: LogOther = attrs.field(factory=LogOther, init=False)
    log_others: LogOthers = attrs.field(factory=LogOthers, init=False)
    log_param: LogParam = attrs.field(factory=LogParam, init=False)
    log_params: LogParams = attrs.field(factory=LogParams, init=False)
    start: Start = attrs.field(factory=Start, init=False)

    @property
    def comet(self) -> comet.CometExperiment:
        return comet.get_running_experiment()  # pyright: ignore[reportReturnType]

    @functools.cached_property
    def exp_dir(self) -> Path:
        return _path.exp_dir(absolute=True)

    @property
    def exp_key(self) -> str:
        return self.comet.get_key()

    @property
    def exp_name(self) -> str:
        return self.comet.get_name()

    @exp_name.setter
    def exp_name(self, value: str) -> None:
        self.comet.set_name(value)

    @property
    def exp_url(self) -> str:
        return self.comet.url  # pyright: ignore[reportReturnType]

    @property
    def project_id(self) -> str:
        return self.comet.project_id  # pyright: ignore[reportReturnType]

    @property
    def project_name(self) -> str:
        return self.comet.project_name

    @functools.cached_property
    def start_time(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone()

    def log_input(self, path: PathLike, /, **kwargs) -> None:
        self.log_asset(path, prefix="inputs/", **kwargs)

    def log_output(self, path: PathLike, /, **kwargs) -> None:
        self.log_asset(path, prefix="outputs/", **kwargs)


exp = Experiment()
add_tag: AddTag = exp.add_tag
add_tags: AddTags = exp.add_tags
end: End = exp.end
log_asset: LogAsset = exp.log_asset
log_code: LogCode = exp.log_code
log_metric: LogMetric = exp.log_metric
log_metrics: LogMetrics = exp.log_metrics
log_other: LogOther = exp.log_other
log_others: LogOthers = exp.log_others
log_param: LogParam = exp.log_param
log_params: LogParams = exp.log_params
start: Start = exp.start



from . import comet, dvc, git, logging
from ._abc import (
    AddTag,
    AddTags,
    End,
    LogAsset,
    LogCode,
    LogMetric,
    LogMetrics,
    LogOther,
    LogOthers,
    LogParam,
    LogParams,
    Plugin,
    Start,
)
from ._exp import (
    Experiment,
    add_tag,
    add_tags,
    end,
    exp,
    log_asset,
    log_code,
    log_metric,
    log_metrics,
    log_other,
    log_others,
    log_param,
    log_params,
    start,
)

__all__ = [
    "AddTag",
    "AddTags",
    "End",
    "Experiment",
    "LogAsset",
    "LogCode",
    "LogMetric",
    "LogMetrics",
    "LogOther",
    "LogOthers",
    "LogParam",
    "LogParams",
    "Plugin",
    "Start",
    "add_tag",
    "add_tags",
    "comet",
    "dvc",
    "end",
    "exp",
    "git",
    "log_asset",
    "log_code",
    "log_metric",
    "log_metrics",
    "log_other",
    "log_others",
    "log_param",
    "log_params",
    "logging",
    "start",
]
